package edu.berkeley.sbp.misc;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.chr.*;
import java.util.*;
import java.io.*;

public class MetaGrammar extends StringWalker {

    public static Union make() throws Exception { return make(MetaGrammarTree.meta, "s"); }
    public static Union make(Tree<String> tree, String nt) throws Exception {
        Meta.MetaGrammarFile mgf = new Meta().new MetaGrammarFile(tree);
        BuildContext bc = new BuildContext(mgf);
        return mgf.get(nt).build(bc);
    }

    ////////////////////////////////////////////////////////////////////////////////

    private static Element  set(Range.Set r) { return CharRange.set(r); }
    private static Element  string(String s) { return CharRange.string(s); }
    /*private*/ static Atom infer(Element e)  { return infer((Topology<Character>)Atom.toAtom(e)); }
    /*private*/ static Atom infer(Topology<Character> t) { return new CharRange(new CharTopology(t)); }

    private MetaGrammar() { }

    public static String string(Iterable<Tree<String>> children) {
        String ret = "";
        for(Tree<String> t : children) ret += string(t);
        return ret;
    }
    public static String string(Tree<String> tree) {
        String ret = "";
        if (tree.head()!=null) ret += tree.head();
        ret += string(tree.children());
        return ret;
    }

    public static class BuildContext extends HashMap<String,Union> {
        private final Meta.MetaGrammarFile mgf;
        public Meta.NonTerminal currentNonTerminal;
        public BuildContext(Meta.MetaGrammarFile mgf) { this.mgf = mgf; }
        public Union build(String s) {
            Union ret = get(s);
            if (ret != null) return ret;
            Meta.NonTerminal mnt = mgf.get(s);
            if (mnt==null) throw new Error("undeclared nonterminal \""+s+"\"");
            return mnt.build(this);
        }
    }

    public static class Meta {
        public Object repeatTag() { return null; }
        public Sequence tryResolveTag(String s, String nonTerminalName, Element[] els, Object[] labels, boolean [] drops) {
            //return resolveTag(s, nonTerminalName, els, labels, drops);
            return null;
        }
        public Sequence resolveTag(String s, String nonTerminalName, Element[] els, Object[] labels, boolean [] drops) {
            return Sequence.rewritingSequence(s, els, labels, drops);
        }
        public class MetaGrammarFile extends HashMap<String,NonTerminal> {
            public MetaGrammarFile(Tree<String> tree) {
                if (!tree.head().equals("grammar")) throw new Error();
                for(Tree<String> nt : tree.child(0))
                    add(new NonTerminal(nt));
            }
            private void add(NonTerminal mnt) {
                if (this.get(mnt.name)!=null) throw new Error("duplicate definition of nonterminal \""+mnt.name+"\"");
                this.put(mnt.name, mnt);
            }
            public String toString() {
                String ret = "";
                for(NonTerminal mnt : this.values()) ret += mnt + "\n";
                return ret;
            }
        }
        public class NonTerminal {
            public String    name;
            public MetaUnion rhs;
            public NonTerminal(Tree<String> tree) {
                name = string(tree.child(0));
                rhs = rhs(tree.child(1));
            }
            public String toString() { return name + " = " + rhs; }
            public Union build(BuildContext bc) {
                NonTerminal ont = bc.currentNonTerminal;
                bc.currentNonTerminal = this;
                try {
                    return rhs.build(bc, name);
                } finally {
                    bc.currentNonTerminal = ont;
                }
            }
        }
        public MetaUnion rhs(Tree<String> t) {
            return t.numChildren()==1
                ? new MetaUnion(t.child(0), false)
                : new MetaUnion(t, true);
        }
        public class MetaUnion implements MetaSequence {
            public boolean prioritized;
            public MetaSequence[] sequences;
            public Sequence buildSequence(BuildContext bc) {
                return Sequence.singleton(new Element[] { buildAnon(bc) }, 0);
            }
            public Union buildAnon(BuildContext bc) {
                String s = "";
                for(int i=0; i<sequences.length; i++)
                    s += (i>0?"\n                "+(prioritized?">":"|")+" ":"")+sequences[i].buildSequence(bc);
                return build(bc, s);
            }
            public Union build(BuildContext bc, String name) {
                Union u = bc.get(name);
                if (u != null) return u;
                u = new Union(name);
                bc.put(name, u);
                HashSet<Sequence> seqs = new HashSet<Sequence>();
                for(MetaSequence s : sequences) {
                    Sequence seq = s.buildSequence(bc);
                    if (seq != null) {
                        Sequence oseq = seq;
                        if (prioritized)
                            for(Sequence seqprev : seqs)
                                seq = seq.not(seqprev);
                        u.add(seq);
                        seqs.add(seq);
                    }
                }
                return u;
            }
            public MetaUnion(Tree<String> t, boolean prioritized) {
                this.prioritized = prioritized;
                int i = 0;
                this.sequences = new MetaSequence[t.numChildren()];
                for(Tree<String> tt : t)
                    sequences[i++] = prioritized
                        ? new MetaUnion(tt, false)
                        : makeMetaSequence(tt);
            }
            public String toString() {
                String ret = "\n     ";
                for(int i=0; i<sequences.length; i++) {
                    ret += sequences[i];
                    if (i<sequences.length-1)
                        ret += (prioritized ? "\n    > " : "\n    | ");
                }
                return ret;
            }
        }

        public interface MetaSequence {
            public abstract Sequence buildSequence(BuildContext bc);
        }

        public MetaSequence makeMetaSequence(Tree<String> t) {
            if ("psx".equals(t.head())) return makeConjunct(t.child(0));
            if (t.head().equals("&"))  return new MetaAnd(makeMetaSequence(t.child(0)), makeConjunct(t.child(1)), false);
            if (t.head().equals("&~")) return new MetaAnd(makeMetaSequence(t.child(0)), makeConjunct(t.child(1)), true);
            return null;
        }

        public class MetaAnd implements MetaSequence {
            boolean not;
            MetaSequence left;
            MetaSequence right;
            public Sequence buildSequence(BuildContext bc) {
                Union u = new Union(toString());
                Sequence ret = left.buildSequence(bc);
                Sequence rs = right.buildSequence(bc);
                rs.lame = true;
                if (not) ret = ret.not(rs);
                else     ret = ret.and(rs);
                u.add(rs);
                u.add(ret);
                return Sequence.singleton(u);
            }
            public MetaAnd(MetaSequence left, MetaSequence right, boolean not) {
                this.left = left;
                this.right = right;
                this.not = not;
            }
            public String toString() { return left + " &"+(not?"~":"")+" "+right; }
        }

        public class Conjunct implements MetaSequence {
            public boolean negated = false;
            public MetaClause[] elements;
            public HashMap<MetaClause,String> labelMap = new HashMap<MetaClause,String>();
            public MetaClause followedBy = null;
            public String tag;
            public MetaClause separator;
            public void addNamedClause(String name, MetaClause mc) {
                labelMap.put(mc, name);
            }
            public Sequence buildSequence(BuildContext bc) {
                Element[] els = new Element[elements.length + (separator==null?0:(elements.length-1))];
                boolean[] drops = new boolean[elements.length + (separator==null?0:(elements.length-1))];
                Object[] labels = new Object[els.length];
                boolean unwrap = false;
                boolean dropAll = false;
                if (tag!=null && tag.equals("[]")) unwrap = true;
                if (tag!=null && "()".equals(tag)) dropAll=true;

                NonTerminal old = bc.currentNonTerminal;
                bc.currentNonTerminal = null;
                for(int i=0; i<elements.length; i++) {
                    int j = separator==null ? i : i*2;
                    els[j] = elements[i].build(bc);
                    labels[j] = labelMap.get(elements[i]);
                    drops[j] = elements[i].drop;
                    if (separator!=null && i<elements.length-1) {
                        els[j+1] = separator.build(bc);
                        drops[j+1] = true;
                    }
                }
                bc.currentNonTerminal = old;

                Sequence ret = null;
                if (dropAll)     ret = Sequence.drop(els, false);
                else if (unwrap) ret = Sequence.unwrap(els, repeatTag(), drops);
                else if (tag!=null) {
                    ret = resolveTag(tag, bc.currentNonTerminal==null ? null : bc.currentNonTerminal.name, els, labels, drops);
                    System.err.println("resolvetag " + tag + " => " + ret);
                } else {
                    ret = tryResolveTag(tag, bc.currentNonTerminal==null ? null : bc.currentNonTerminal.name, els, labels, drops);
                    if (ret==null) {
                        int idx = -1;
                        for(int i=0; i<els.length; i++)
                            if (!drops[i])
                                if (idx==-1) idx = i;
                                else throw new Error("multiple non-dropped elements in sequence: " + Sequence.drop(els,false));
                        if (idx != -1) ret = Sequence.singleton(els, idx);
                        else           ret = Sequence.drop(els, false);
                    }
                }
                if (this.followedBy != null)
                    ret.follow = infer(this.followedBy.build(bc));
                return ret;
            }
            private Conjunct(Tree<String> t) {
                elements = new MetaClause[t.numChildren()];
                int i = 0; for(Tree<String> tt : t)
                    elements[i++] = makeMetaClause(tt, this);
            }
            public String toString() {
                String ret = (tag==null ? "" : (tag+":: ")) + (negated ? "~" : "");
                if (elements.length > 1) ret += "(";
                for(MetaClause mc : elements) ret += (" " + mc + " ");
                if (separator != null) ret += " /" + separator;
                if (followedBy != null) ret += " -> " + followedBy;
                if (elements.length > 1) ret += ")";
                return ret;
            }
        }
            public Conjunct makeConjunct(Tree<String> t) {
                //System.err.println("makeConjunct("+t+")");
                if ("/".equals(t.head())) {
                    Conjunct ret = makeConjunct(t.child(0));
                    ret.separator = makeMetaClause(t.child(1), ret);
                    return ret;
                }
                if ("->".equals(t.head())) {
                    Conjunct ret = makeConjunct(t.child(0));
                    ret.followedBy = makeMetaClause(t.child(1), ret);
                    return ret;
                }
                if ("::".equals(t.head())) {
                    Conjunct ret = makeConjunct(t.child(1));
                    ret.tag = string(t.child(0));
                    return ret;
                }
                if ("ps".equals(t.head())) {
                    return new Conjunct(t.child(0));
                }
                return new Conjunct(t);
            }

            public MetaClause makeMetaClause(Tree<String> t, Conjunct c) {
                //System.err.println("MetaClause.makeMetaClause("+t+")");
                if (t==null) return new Epsilon();
                if (t.head()==null) return new Epsilon();
                if (t.head().equals("{")) return new MetaTree(makeConjunct(t.child(0)));
                if (t.head().equals("*")) return new MetaRepeat(makeMetaClause(t.child(0), c), false, null, true, true);
                if (t.head().equals("+")) return new MetaRepeat(makeMetaClause(t.child(0), c), false, null, false, true);
                if (t.head().equals("?")) return new MetaRepeat(makeMetaClause(t.child(0), c), false, null, true, false);
                if (t.head().equals("**")) return new MetaRepeat(makeMetaClause(t.child(0), c), true, null, true, true);
                if (t.head().equals("++")) return new MetaRepeat(makeMetaClause(t.child(0), c), true, null, false, true);
                if (t.head().equals("*/")) return new MetaRepeat(makeMetaClause(t.child(0), c), false, makeMetaClause(t.child(1), c), true, true);
                if (t.head().equals("+/")) return new MetaRepeat(makeMetaClause(t.child(0), c), false, makeMetaClause(t.child(1), c), false, true);
                if (t.head().equals("**/")) return new MetaRepeat(makeMetaClause(t.child(0), c), true, makeMetaClause(t.child(1), c), true, true);
                if (t.head().equals("++/")) return new MetaRepeat(makeMetaClause(t.child(0), c), true, makeMetaClause(t.child(1), c), false, true);
                if (t.head().equals("()")) return new Epsilon();
                if (t.head().equals("[")) return new MetaRange(t.child(0));
                if (t.head().equals("literal")) return new StringLiteral(t.child(0));
                if (t.head().equals("nonTerminal")) return new NonTerminalReference(t.child(0));
                if (t.head().equals(")")) return new SelfReference();
                if (t.head().equals("(")) return new Parens(t.child(0));
                if (t.head().equals("~")) return new MetaInvert(t.child(0), c);
                if (t.head().equals("!")) { MetaClause mc = makeMetaClause(t.child(0), c); mc.drop = true; return mc; }
                if (t.head().equals("^")) { c.tag = string(t.child(0)); return new StringLiteral(t.child(0)); }
                if (t.head().equals("^^")) throw new Error("carets: " + t);
                if (t.head().equals(":")) {
                    String name = string(t.child(0));
                    MetaClause clause = makeMetaClause(t.child(1), c);
                    c.addNamedClause(name, clause);
                    return clause;
                }
                throw new Error("unknown: " + t);
            }

        public abstract class MetaClause {
            public String label = null;
            public boolean drop = false;
            public boolean lift = false;
            public abstract Element build(BuildContext bc);
        }
            public class MetaRepeat extends MetaClause {
                public MetaClause element, separator;
                public boolean maximal, zero, many;
                public Element build(BuildContext bc) {
                    return !maximal
                        ? (separator==null
                           ? Sequence.repeat(element.build(bc), zero, many, null, repeatTag())
                           : Sequence.repeat(element.build(bc), zero, many, separator.build(bc), repeatTag()))
                        : (separator==null
                           ? Sequence.repeatMaximal(infer(element.build(bc)), zero, many, repeatTag())
                           : Sequence.repeatMaximal(element.build(bc), zero, many, infer(separator.build(bc)), repeatTag()));
                }
                public MetaRepeat(MetaClause element, boolean maximal, MetaClause separator, boolean zero, boolean many) {
                    this.separator = separator;
                    this.element = element;
                    this.maximal = maximal;
                    this.zero = zero;
                    this.many = many;
                }
                public String toString() {
                    return element+
                        ((zero&&!many)?"?":zero?"*":"+")+
                        (!maximal?"":zero?"*":"+")+
                        (separator==null?"":(" /"+separator));
                }
            }
        public class Epsilon extends MetaClause {
            public String toString() { return "()"; }
            public Element build(BuildContext bc) { return Union.epsilon; }
        }
            public class Parens extends MetaClause {
                public MetaUnion body;
                public Parens(Tree<String> t) { this.body = rhs(t); }
                public String toString() { return "( " + body + " )"; }
                public Element build(BuildContext bc) { return body.buildAnon(bc); }
            }

            public class MetaTree extends MetaClause {
                public Conjunct body;
                public MetaTree(Conjunct c) { this.body = c; }
                public String toString() { return "{ " + body + " }"; }
                public Element build(BuildContext bc) {
                    Union u = new Union();
                    Union u2 = new Union();
                    Sequence seq = body.buildSequence(bc);
                    u2.add(seq);
                    u.add(Sequence.singleton(new Element[] { CharRange.leftBrace,
                                                             new NonTerminalReference("ws").build(bc),
                                                             u2,
                                                             new NonTerminalReference("ws").build(bc),
                                                             CharRange.rightBrace }
                                             , 2));
                    //u.add(seq);
                    return u;
                }
            }

            public class MetaRange extends MetaClause {
                Range.Set range = new Range.Set();
                public String toString() { return range.toString(); }
                public Element build(BuildContext bc) { return set(range); }
                public MetaRange(Tree<String> t) {
                    for(Tree<String> tt : t) {
                        if (tt.head().equals("range")) {
                            range.add(tt.child(0).head().charAt(0));
                        } else if (tt.head().equals("-")) {
                            range.add(new Range(string(tt.child(0)).charAt(0),
                                                string(tt.child(1)).charAt(0)));
                        }
                    }
                }
            }
            public class StringLiteral extends MetaClause {
                public String literal;
                public Element build(BuildContext bc) { return string(literal); }
                public StringLiteral(Tree<String> literal) { this.literal = string(literal); this.drop = true; }
                public String toString() { return "\""+StringUtil.escapify(literal, "\"\r\n\\")+"\""; }
            }
            public class NonTerminalReference extends MetaClause {
                public String name;
                public NonTerminalReference(Tree<String> name) { this.name = string(name); }
                public NonTerminalReference(String name) { this.name = name; }
                public Element build(BuildContext bc) { return bc.build(name); }
                public String toString() { return name; } 
            }
            public class SelfReference extends MetaClause {
                public String toString() { return "(*)"; } 
                public Element build(BuildContext bc) { return new Union("(*)"); /* FIXME */ }
            }
            public class MetaInvert extends MetaClause {
                public MetaClause element;
                public MetaInvert(Tree<String> t, Conjunct c) { this.element = makeMetaClause(t, c); }
                public String toString() { return "~"+element; }
                public Element build(BuildContext bc) { return infer((Topology<Character>)Atom.toAtom(element.build(bc)).complement()); }
            }

    }

    public static void main(String[] args) throws Exception {
        if (args.length != 2) {
            System.err.println("usage: java " + MetaGrammar.class.getName() + " grammarfile.g com.yourdomain.package.ClassName");
            System.exit(-1);
        }
        //StringBuffer sbs = new StringBuffer();
        //((MetaGrammar)new MetaGrammar().walk(meta)).nt.get("e").toString(sbs);
        //System.err.println(sbs);
        String className   = args[1].substring(args[1].lastIndexOf('.')+1);
        String packageName = args[1].substring(0, args[1].lastIndexOf('.'));
        String fileName    = packageName.replace('.', '/') + "/" + className + ".java";

        BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(fileName)));
        StringBuffer out = new StringBuffer();

        boolean skip = false;
        for(String s = br.readLine(); s != null; s = br.readLine()) {
            if (s.indexOf("DO NOT EDIT STUFF BELOW: IT IS AUTOMATICALLY GENERATED") != -1 && s.indexOf("\"")==-1) skip = true;
            if (s.indexOf("DO NOT EDIT STUFF ABOVE: IT IS AUTOMATICALLY GENERATED") != -1 && s.indexOf("\"")==-1) break;
            if (!skip) out.append(s+"\n");
        }

        out.append("\n        // DO NOT EDIT STUFF BELOW: IT IS AUTOMATICALLY GENERATED\n");
        Tree<String> ts = new CharParser(MetaGrammar.make()).parse(new FileInputStream(args[0])).expand1();

        //Forest<String> fs = new CharParser(make()).parse(new FileInputStream(args[0]));
        //System.out.println(fs.expand1());

        //GraphViz gv = new GraphViz();
        //fs.toGraphViz(gv);
        //FileOutputStream fox = new FileOutputStream("out.dot");
        //gv.dump(fox);
        //fox.close();

        ts.toJava(out);
        out.append("\n        // DO NOT EDIT STUFF ABOVE: IT IS AUTOMATICALLY GENERATED\n");

        for(String s = br.readLine(); s != null; s = br.readLine()) out.append(s+"\n");
        br.close();

        OutputStream os = new FileOutputStream(fileName);
        PrintWriter p = new PrintWriter(new OutputStreamWriter(os));
        p.println(out.toString());
        p.flush();
        os.close();
    }

}
